# -*- python -*-
# $RCSfile: gfxwindow.py,v $
# $Revision: 1.145 $
# $Author: langer $
# $Date: 2004/09/22 19:05:10 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from gtk import *

import math, time
from SWIG.common.coord import Coord
from SWIG.common import lock
from SWIG.common import switchboard
from SWIG.common.IO.GUI import oofcanvas
from SWIG.common.IO.GUI import rubberband
from common import debug
from common import mainthread
from common import primitives
from common import subthread
from common import toolbox
from common import utils
from common.IO import ghostgfxwindow
from common.IO import gfxmanager
from common.IO import layereditor
from common.IO import mainmenu
from common.IO import reporter
from common.IO.GUI import chooser
from common.IO.GUI import fileselector
from common.IO.GUI import gfxmenu
from common.IO.GUI import gtkutils
from common.IO.GUI import layerwidget
from common.IO.GUI import mousehandler
from common.IO.GUI import quit
from common.IO.GUI import scrollcanvas
from common.IO.GUI import subWindow
from common.IO.GUI import toolboxGUI
from common.guitop import top
from types import *
import common.color         # preserve 'color' for a local var name.


BUFFERING = 1 # whether to use BufferedOutputDevice or unbuffered CanvasOutput

# during_callback() is called (by CanvasOutput.show()) only in
# non-threaded mode, so we don't worry about the thread-safety of a
# global variable here.
_during_callback = 0
def during_callback():
    return _during_callback

_pi_over_2 = math.pi/2.0

class ContourMapData:
    # This class just aggregates data related to the ContourMap
    # display, to keep the code tidy.
    def __init__(self):
        self.canvas = None
        self.device = None
        self.mouse_down = None
        self.mark_value = None
        self.canvas_mainlayer = None
        self.canvas_ticklayer = None

class GfxWindow(subWindow.SubWindow, ghostgfxwindow.GhostGfxWindow):
    # TODO LATER: Set initial sizes in some more user-friendly way.
    initial_height = 400
    initial_width = 800
    gutter_size = 10
    def __init__(self, name, gfxmanager, clone=0):
        mainthread.runBlock(self.__init__thread, (name, gfxmanager, clone))

    def __init__thread(self, name, gfxmanager, clone):
        self.name = name

        self.oofcanvas = None
        self.realized = 0
        self.zoomed = 0
        self.settings = ghostgfxwindow.GfxSettings()
        self.mouseHandler = mousehandler.nullHandler # doesn't do anything
        self.rubberband = rubberband.NoRubberBand()

        self.contourmapdata = ContourMapData()

        self.gfxlock = lock.Lock() ## lock to guarantee that each gfxwindow
                                   ## will only have one thread.
        # Build all the GTK objects for the interior of the box.  These
        # actually get added to the window itself after the SubWindow
        # __init__ call.  They need to be created first so the
        # GhostGfxWindow can operate on them, and then create the menus
        # which are handed off to the SubWindow.
        self.mainpane = GtkVPaned()
        self.mainpane.set_gutter_size(self.gutter_size)

        # Panes dividing upper pane horizontally into 3 parts.
        # paned1's left half contains paned2.
        self.paned1 = GtkHPaned()
        self.mainpane.pack1(self.paned1,resize=TRUE)
        self.paned1.set_gutter_size(self.gutter_size)

        # paned2 is in left half of paned1
        self.paned2 = GtkHPaned()
        self.paned1.pack1(self.paned2, resize=TRUE)
        self.paned2.set_gutter_size(self.gutter_size)

        # The toolbox is in the left half of paned2 (ie the left frame of 3)
        toolboxframe = GtkFrame()
        toolboxframe.set_shadow_type(SHADOW_IN)
        self.paned2.pack1(toolboxframe, resize=TRUE)

        # Box containing the toolbox label and the scroll window for
        # the toolbox itself.
        toolboxbox1 = GtkVBox()
        toolboxframe.add(toolboxbox1)
        hbox = GtkHBox()
##        hbox.set_usize(50, -1)
        toolboxbox1.add(hbox, expand=0, fill=0)
        hbox.add(GtkLabel("Toolbox:"), expand=0, fill=0, padding=3)
        
        self.toolboxchooser = chooser.ChooserWidget([],
                                                    callback=self.switchToolbox)
        hbox.add(self.toolboxchooser.gtk, expand=0, fill=0, padding=3)

        # Scroll window for the toolbox itself.
        toolboxbox2 = GtkScrolledWindow()
        toolboxbox2.set_policy(POLICY_AUTOMATIC, POLICY_AUTOMATIC)
        toolboxbox1.add(toolboxbox2, expand=1, fill=1)

        # Actually, the tool box goes inside yet another box, so that
        # we have a GtkVBox that we can refer to later.
        self.toolboxbody = GtkVBox()
        toolboxbox2.add_with_viewport(self.toolboxbody)

        self.toolboxGUIs = []           # GUI wrappers for toolboxes.
        self.current_toolbox = None

        # The canvas is in the right half of paned2 (ie the middle pane of 3)
        self.newCanvas()
        
        # the contourmap is in the right half of paned1 (the right pane of 3)
        contourmapframe = GtkFrame()
        contourmapframe.set_shadow_type(SHADOW_IN)
        self.paned1.pack2(contourmapframe, resize=FALSE)

        contourmapbox = GtkVBox()
        contourmapframe.add(contourmapbox)

        contourmap_max_box = GtkHBox()
        self.contourmap_max = GtkLabel("max")
        contourmap_max_box.add(self.contourmap_max)
        self.new_contourmap_canvas()    # Sets self.contourmapdata.canvas.
        contourmap_min_box = GtkHBox()        
        self.contourmap_min = GtkLabel("min")
        contourmap_min_box.add(self.contourmap_min)
        
        contourmaplevelbox = GtkHBox()
##         contourmaplevelbox.pack_start(GtkLabel(" Mark "), FALSE, FALSE)
##         contourmaplevelbox.pack_start(GtkVSeparator(), FALSE, FALSE)
        self.contourlevel_min = GtkLabel("   min   ")
        contourmaplevelbox.pack_start(self.contourlevel_min, TRUE, TRUE)
        contourmaplevelbox.pack_start(GtkVSeparator(), FALSE, FALSE)
        contourmaplevelbox.pack_start(GtkVSeparator(), FALSE, FALSE)
        self.contourlevel_max = GtkLabel("   max   ")
        contourmaplevelbox.pack_end(self.contourlevel_max, TRUE, TRUE)
        
        contourmapclearbutton = GtkButton("Clear Mark")
        contourmapclearbutton.connect("clicked", self.contourmap_clear_marker)
        
        contourmapbox.pack_start(contourmap_max_box, FALSE, FALSE)
        contourmapbox.pack_start(GtkHSeparator(), FALSE, FALSE)
        self.contourmapdata.canvas.add_self(contourmapbox._o)
        contourmapbox.pack_start(GtkHSeparator(), FALSE, FALSE)
        contourmapbox.pack_start(contourmap_min_box, FALSE, FALSE)

        contourmapbox.pack_start(GtkHSeparator(), FALSE, FALSE)
        contourmapbox.pack_start(GtkHSeparator(), FALSE, FALSE)
        contourmapbox.pack_start(contourmaplevelbox, FALSE, FALSE)
        contourmapbox.pack_start(GtkHSeparator(), FALSE, FALSE)
        contourmapbox.pack_end(contourmapclearbutton, FALSE, FALSE)

        contourmapframe.show_all()


        # HACK.  Set the position of the toolbox/canvas divider.  This
        # prevents the toolbox pane from coming up minimized.
        self.paned2.set_position(250)

        # Bottom part of main pane is a list of layers.  The actual
        # DisplayLayer objects are stored in self.display.
        # self.layerList is just the Gtk representation of the list.
        layerFrame = GtkFrame(label='Layers')
        self.mainpane.pack2(layerFrame, resize=FALSE)
        self.innerLayerFrame = GtkScrolledWindow()
        self.innerLayerFrame.set_policy(POLICY_AUTOMATIC, POLICY_AUTOMATIC)
        layerFrame.add(self.innerLayerFrame)
        # The list is really a list of LayerWidgets.  We have to know
        # which widget corresponds to which layer.  The layerwidgets
        # dictionary does this for us.  We *don't* just stick a
        # 'widget' attribute into the DisplayMethod object because
        # that would mess up the equality check for DisplayMethods.
        # print "Clearing layerwidgets dict.\n"
        self.layerwidgets = {}
        self.layerList = None
        self.makeLayerList()

        # The GhostGfxWindow base class __init__ is finally called
        # here.  It's not called earlier because it calls some
        # functions which are redefined in GfxWindow, and which
        # operate on the gtk objects.
        ghostgfxwindow.GhostGfxWindow.__init__(self, name, gfxmanager,
                                               clone=clone)

        # Add gui callbacks to the non-gui menu created by the GhostGfxWindow.
        filemenu = self.menu.File
        filemenu.Save_Image.gui_callback = self.saveImage_gui
        filemenu.Save_Contourmap.gui_callback = self.saveContourmap_gui
        filemenu.Quit.gui_callback = quit.quit
        layermenu = self.menu.Layer
        layermenu.New.gui_callback = self.newLayer_gui
        layermenu.Edit.gui_callback = self.editLayer_gui
        layermenu.Delete.gui_callback = self.deleteLayer_gui
        layermenu.Hide.gui_callback = self.hideLayer_gui
        layermenu.Show.gui_callback = self.showLayer_gui
        layermenu.Hide_Contour_Map.gui_callback = self.hideLayerContourmap_gui
        layermenu.Show_Contour_Map.gui_callback = self.showLayerContourmap_gui
        layermenu.Raise.One_Level.gui_callback = self.raiseLayer_gui
        layermenu.Raise.To_Top.gui_callback = self.raiseToTop_gui
        layermenu.Lower.One_Level.gui_callback = self.lowerLayer_gui
        layermenu.Lower.To_Bottom.gui_callback = self.lowerToBottom_gui
        settingmenu = self.menu.Settings
        toolboxmenu = self.menu.Toolbox

        # Construct gui's for toolboxes.  This must be done after the
        # base class is constructed so that all non-gui toolboxes are
        # created before any of their gui versions.  Some gui
        # toolboxes need to know about more than one non-gui toolbox.
        map(self.makeToolboxGUI, self.toolboxes)

        # Raise_window routine is in SubWindow class.
##        self.menu.gui_callback = self.raise_window
        getattr(mainmenu.OOF.Windows.Graphics, name).gui_callback \
                                               = self.raise_window


        # SubWindow initializer makes the menu bar, and sets up the
        # .gtk and .mainbox members.  ".gtk" is the window itself,
        # and .mainbox is a GtkVBox that holds the menu bar.
        windowname = utils.underscore2space("OOF2 " + name)
        subWindow.SubWindow.__init__(
            self, windowname, menu=self.menu)

        # TODO LATER: This whole initialization sequence is a little
        # out of control.  The issue is that the GhostGfxWindow class
        # calls redefined functions which operate on some of the
        # widgets, which therefore must exist, but it itself creates
        # the menu, which SubWindow requires in order to construct the
        # menu bar.  This arrangement works, but the relatively long
        # lifetime of disconnected gtk widgets, along with the
        # deferral of the subclass inits, is clumsy.  The
        # initialization should be reviewed with an eye to
        # rationalization.

        self.gtk.connect('destroy', self.destroyCB)
        self.gtk.connect_after('realize', self.realizeCB)
        self.gtk.set_usize(GfxWindow.initial_width, GfxWindow.initial_height)
        self.gtk.set_policy(1, 1, 0)

        self.mainbox.set_homogeneous(0)
        self.mainbox.set_spacing(3)

        self.mainbox.add(self.mainpane, fill=1, expand=1)
        
        ## self.draw()
        self.gtk.show_all()

        self.updateToolboxChooser()

    def __repr__(self):
        return 'GfxWindow("%s")' % self.name

    ################################################

    def newCanvas(self):
        mainthread.runBlock(self.newCanvas_thread)
    def newCanvas_thread(self):
        debug.mainthreadTest()
        ppu = None
        if self.oofcanvas:
            ppu = self.oofcanvas.get_pixels_per_unit()
            scrollregion = self.oofcanvas.get_scrollregion()
            self.oofcanvas.destroy()

        self.oofcanvas = scrollcanvas.ScrollCanvas(
            antialias=self.settings.antialias,
            hScrollCB=self.hCanvasScrollCB,
            vScrollCB=self.vCanvasScrollCB)

        if ppu is not None:
            self.oofcanvas.set_pixels_per_unit(ppu)
            self.oofcanvas.set_scrollregion(scrollregion)

        self.paned2.pack2(self.oofcanvas, resize=TRUE)

        # delayed import to avoid import loops
        from common.IO.GUI import canvasoutput
        from common.IO import outputdevice
        if BUFFERING:
            self.device = outputdevice.BufferedOutputDevice(
                canvasoutput.CanvasOutput(self.oofcanvas))
        else:
            self.device = canvasoutput.CanvasOutput(self.oofcanvas)

        # On a new canvas, these should both be set at once,
        # since they both need to call underlay.
        self.oofcanvas.set_underlay_params(
            self.settings.bgcolor, self.settings.margin)
        # self.oofcanvas.set_margin(self.settings.margin)
        # self.oofcanvas.set_bgColor(self.settings.bgcolor)
        self.oofcanvas.set_mouse_callback(self.mouseCB)
        self.oofcanvas.set_rubberband(self.rubberband)
        self.oofcanvas.show()

    def get_oofcanvas(self):
        return self.oofcanvas

    def newLayerMembers(self):
        self.fillLayerList()
        ghostgfxwindow.GhostGfxWindow.newLayerMembers(self)


    def makeLayerList(self):
        mainthread.runBlock(self.makeLayerList_thread)
    def makeLayerList_thread(self):
        # Create or recreate the Gtk list that holds the layers.
        # Don't fill it yet, since this can be called before the
        # display is set up.
        # Anesthetize the list before destroying it, or else a
        # selected item will emit a 'deselect' signal when it's
        # removed.
        if self.layerList is not None:
            self.layerList.set_sensitive(0)
            self.layerList.destroy()

        # print "Clearing layerwidgets dict.\n"
        self.layerwidgets = {}

        self.layerList = GtkList()
        self.innerLayerFrame.add_with_viewport(self.layerList)
        self.layerList.show_all()
        
    def fillLayerList(self):
        mainthread.runBlock(self.fillLayerList_thread)
    def fillLayerList_thread(self):
        # Put the layers in their list.  The list has to be explicitly
        # reconstructed when layers are deleted, raised, or lowered.
        # (It is theoretically possible to avoid reconstructing the
        # whole list, but it's not worth the effort, especially
        # because the list seems to be confused about what's selected
        # when it gets reconstructed.  So it's simpler to get rid of
        # it and start all over.)

        self.makeLayerList()
        # Create new layer widgets, prepending them to the list so
        # that the top layer (last in the drawing order) is at the top
        # of the list.
        count = 0
        for layer in self.display:
            widget = self.makeLayerWidget(layer)
            if count == 0:
                widget.disableLowerButton() # can't lower the bottom layer
            self.layerList.prepend_items([widget.gtk])
            if layer is self.selectedLayer:
                widget.select()
            count += 1
        if count > 0:
            widget.disableRaiseButton() # can't raise the top layer
                    
        self.layerList.show_all()
        # Notify the color-bar pane.
        self.contourmap_newlayers()

    # The layerwidgets dictionary is maintained by the widgets through
    # the following calls:
    def addLayerWidget(self, layer, widget):
        # Called by LayerWidget.__init__(), and by LayerWidget.replaceMethod().
        # print "Adding widget ", widget, "indexed by layer ", id(layer), str(layer)[:20]
        self.layerwidgets[layer] = widget
    def removeLayerWidget(self, layer):
        # print "Removing widget indexed by layer ", layer
        # Called from the gtk destroy callback in a LayerWidget,
        # and by LayerWidget.replaceMethod().
        del self.layerwidgets[layer]

    def replaceLayer(self, count, oldlayer, newlayer):
        self.gfxlock.acquire()
        try:
            ghostgfxwindow.GhostGfxWindow.replaceLayer(self, count,
                                                       oldlayer, newlayer)
            mainthread.runBlock(
                self.layerwidgets[oldlayer].replaceMethod, (newlayer,))
        finally:
            self.gfxlock.release()

    ###########################################
    # Color bar stuff.

    # Create object and assign to self.contourmap_canvas.
    def new_contourmap_canvas(self):
        mainthread.runBlock(self.new_contourmap_canvas_thread)
    def new_contourmap_canvas_thread(self):
        if self.contourmapdata.canvas:
            self.contourmapdata.canvas.destroy()
        # Argument is antialias parameter.
        self.contourmapdata.canvas = oofcanvas.OOFCanvas(0)

        from common.IO.GUI import canvasoutput
        self.contourmapdata.device=canvasoutput.CanvasOutput(
            self.contourmapdata.canvas)

        self.contourmapdata.canvas.set_configure_callback(
            self.contourmap_config)
        self.contourmapdata.canvas.set_mouse_callback(self.contourmap_mouse)
        # Params are background color and margin.
        self.contourmapdata.canvas.set_underlay_params(
            self.settings.bgcolor, 0.0)

        # Create two layers, one for the "main" drawing, and
        # one for the ticks.
        self.contourmapdata.canvas_mainlayer = \
                                         self.contourmapdata.canvas.newLayer()
        self.contourmapdata.canvas_ticklayer = \
                                         self.contourmapdata.canvas.newLayer()
        self.contourmapdata.canvas_ticklayer.raise_to_top()
    
    # Function called after layers have been arranged.  Does not
    # draw the actual contourmap, just records the layers.
    def contourmap_newlayers(self):
        ghostgfxwindow.GhostGfxWindow.contourmap_newlayers(self)
        for ell in self.display:
            if ell==self.current_contourmap_method:
                mainthread.runBlock(
                    self.layerwidgets[ell].setContourmapButton,(1,))
            else:
                mainthread.runBlock(
                    self.layerwidgets[ell].setContourmapButton,(0,))

    # def set_contourmap_layer(self, method):
    #     ghostgfxwindow.GhostGfxWindow.set_contourmap_layer(self, method)
        
            
    # Draw the contourmap onto the canvas, and update the numbers on the
    # display.  Called after all the layers have been drawn in the
    # main pane.  Also called from the menu callback which resets the
    # aspect ratio.
    def show_contourmap_info(self):
        if not self.gtk:
            return

        self.contourmapdata.canvas_mainlayer.clear()
        self.contourmapdata.canvas_mainlayer.make_current()
        
        c_min = None
        c_max = None
        if self.current_contourmap_method:
            self.current_contourmap_method.draw_contourmap(
                self, self.contourmapdata.device)
            (c_min, c_max, lvls) = \
                    self.current_contourmap_method.get_contourmap_info()
            # Zoom and scroll the canvas so that the drawn contour map 
            # exactly fills it vertically.
            ph = self.contourmapdata.canvas.get_height_in_pixels()
            # c_max can equal c_min....
            if c_max!=c_min:
                ppu = 1.0*ph/abs(c_max-c_min)
            else:
                ppu = ph  # Arbitrary, one unit for all the pixels.
                
            self.contourmapdata.canvas.set_pixels_per_unit(ppu)

            # print self.contourmapdata.canvas.get_bounds()
            self.contourmapdata.canvas.set_scrollregion(
                self.contourmapdata.canvas.get_bounds())

        self.contourmapdata.device.end_layer()
        self.contourmapdata.device.show()
        
        if c_min is None:
            self.contourmap_min.set_text('min')
        else:
            self.contourmap_min.set_text("%.3e" % c_min)
            
        if c_max is None:
            self.contourmap_max.set_text('max')
        else:
            self.contourmap_max.set_text("%.3e" % c_max)

        self.show_contourmap_ticks(self.contourmapdata.mark_value)


    # Draw the marks on it.  Argument is the new value for the ticks.
    # A tick-layer redraw can be forced by setting
    # self.contourmapdata.mark_value to None and then calling this with
    # the new value.
    def show_contourmap_ticks(self, y):
        if not self.gtk:
            return
        c_min = None
        c_max = None
        
        if self.current_contourmap_method:
            (c_min, c_max, lvls) = \
                    self.current_contourmap_method.get_contourmap_info()

        if ((c_max is not None) and (c_min is not None)):
            if y is not None:
                real_y = y + c_min
                # need to know which level am I in.
                level = None
                for i in range(len(lvls)):
                    if real_y <= lvls[i]:
                        level = i-1
                        if i==len(lvls):
                            i -= 1
                        break
                if level is None:
                    self.contourmapdata.canvas_ticklayer.clear()
                    self.contourlevel_min.set_text('')
                    self.contourlevel_max.set_text('')
                else:
                    self.contourmapdata.canvas_ticklayer.destroy()
                    self.contourmapdata.canvas_ticklayer = self.contourmapdata.canvas.newLayer()
                    self.contourmapdata.canvas_ticklayer.raise_to_top()
                    self.contourmapdata.canvas_ticklayer.make_current()
                
                    self.contourmapdata.device.set_lineColor(
                        self.settings.contourmap_markercolor)
                    self.contourmapdata.device.set_lineWidth(
                        self.settings.contourmap_markersize)
                    width=(c_max-c_min)/self.settings.aspectratio

                    polygon = primitives.Polygon(
                        [primitives.Point(0.0, lvls[level]-c_min),
                         primitives.Point(width, lvls[level]-c_min),
                         primitives.Point(width, lvls[level+1]-c_min),
                         primitives.Point(0.0, lvls[level+1]-c_min)
                         ])
                    self.contourmapdata.device.draw_polygon(polygon)

                    self.contourmapdata.mark_value = y
                    minimum = lvls[level]
                    maximum = lvls[level] + (lvls[1]-lvls[0])
                    self.contourlevel_min.set_text("%.2e" % minimum)
                    self.contourlevel_max.set_text("%.2e" % maximum)
            else:
                self.contourmapdata.canvas_ticklayer.clear()
                self.contourlevel_min.set_text('')
                self.contourlevel_max.set_text('')

    # Callback for size changes of the pane containing the contourmap.
    def contourmap_config(self, x, y, width, height):
        if self.current_contourmap_method:
            (c_min, c_max, c_lvls) = \
                    self.current_contourmap_method.get_contourmap_info()
            if c_max!=c_min:
                ppu = 1.0*height/(c_max-c_min)
                # ppu = 1.0*height/1000.0
            else:
                ppu = height # Arbitrary

            self.contourmapdata.canvas.set_pixels_per_unit(ppu)
            self.contourmapdata.canvas.set_scrollregion(
                self.contourmapdata.canvas.get_bounds())

    def contourmap_mouse(self, event, x, y, shift, ctrl):
        if event=="down":
            self.contourmapdata.mouse_down = 1
        elif (event=="move") and self.contourmapdata.mouse_down==1:
            mainthread.run(self.show_contourmap_ticks, (y,))
        elif event=="up":
            self.contourmapdata.mouse_down = None
            mainthread.run(self.show_contourmap_ticks, (y,))



    # Button callback.
    def contourmap_clear_marker(self, gtk):
        self.contourmapdata.mark_value = None
        self.contourlevel_min.set_text('')
        self.contourlevel_max.set_text('')
        mainthread.run(self.show_contourmap_info)


    # Overridden menu callbacks for the contourmap show/hide menu item.
    def hideLayerContourmap(self, menuitem, n):
        ghostgfxwindow.GhostGfxWindow.hideLayerContourmap(self, menuitem, n)
        layer = self.display[n]
        mainthread.runBlock(self.layerwidgets[layer].setContourmapButton, (0,))
        # self.contourmap_newlayers()
        self.show_contourmap_info()
        
    def showLayerContourmap(self, menuitem, n):
        mainthread.runBlock(self.showLayerContourmap_thread, (menuitem, n))
    def showLayerContourmap_thread(self, menuitem, n):
        ghostgfxwindow.GhostGfxWindow.showLayerContourmap(self, menuitem, n)
        layer = self.display[n]
        for ell in self.display:
            self.layerwidgets[ell].setContourmapButton(0)
        self.layerwidgets[layer].setContourmapButton(1)
        # self.contourmap_newlayers()
        self.show_contourmap_info()


    
    # GUI callbacks -- required because the GUI menu item must
    # operate on the current layer.
    def hideLayerContourmap_gui(self, menuitem):
        if self.selectedLayer is None:
            reporter.report(
                "Unable to hide contour map, no layer is selected.")
        else:
            self.hideLayerContourmap(menuitem,
                                     self.layerID(self.selectedLayer))

    def showLayerContourmap_gui(self, menuitem):
        if self.selectedLayer is None:
            reporter.report(
                "Unable to show contour map, no layer is selected.")
        else:
            self.showLayerContourmap(menuitem,
                                     self.layerID(self.selectedLayer))

    ###########################################
   
    def realizeCB(self, *args):         # gtk callback
        if not self.realized:
            self.realized = 1
##            width = self.mainpane.get_allocation()[2]
##            height = self.mainpane.get_allocation()[3]
##            self.mainpane.set_position(2.*height/3.)
##            self.paned1.set_position(4.*width/5.)
##            self.paned2.set_position(width/5.)
            if not self.zoomed:
                self.zoomFillWindow()
        return FALSE
        
    def close(self, *args):             # OOFMenu callback
        ghostgfxwindow.GhostGfxWindow.close(self, *args)
        for tbgui in self.toolboxGUIs:
            if tbgui.active:
                tbgui.deactivate()
            tbgui.close()
        del self.toolboxGUIs
        del self.mouseHandler
        if self.gtk:
            mainthread.runBlock(self.gtk.destroy) # calls destroyCB via gtk

    def destroyCB(self, *args):         # gtk callback
        self.device.destroy() ## tell all the miniThreads to stop and go home.
        # See comment in GhostGfxWindow.close about the order of operations.
        if self.gtk:
            if self.menu:
                self.menu.gui_callback=None
            self.gtk = None             # make sure this isn't repeated
            if self.menu:
                self.menu.File.Close()    # calls self.close()
            self.oofcanvas.destroy()
            del self.oofcanvas
            self.contourmapdata.device.destroy()
            del self.contourmapdata.device
            ## del self.device

##################################################

    def switchToolbox(self, chooser, tbname): # toolboxchooser callback
        self.selectToolbox(tbname)

    def selectToolbox(self, tbname):
        debug.mainthreadTest()
        if not (self.current_toolbox and (self.current_toolbox.name()==tbname)):
            self.removeMouseHandler()
            if self.current_toolbox:
                self.current_toolbox.deactivate()
            self.current_toolbox = self.getToolboxGUIByName(tbname)
            self.current_toolbox.activate()
            for c in self.toolboxbody.children():
                self.toolboxbody.remove(c)
            self.toolboxbody.add(self.current_toolbox.gtk)
            self.toolboxbody.show_all()

    def getToolboxGUIByName(self, name):
        for tbgui in self.toolboxGUIs:
            if tbgui.name() == name:
                return tbgui

    def makeToolboxGUI(self, tb):
        tbgui = mainthread.runBlock(tb.makeGUI)
        if tbgui:
            self.toolboxGUIs.append(tbgui)
            self.toolboxGUIs.sort()
            self.updateToolboxChooser()
            if len(self.toolboxGUIs)==1:
                self.selectToolbox(tbgui.name())
               
    def updateToolboxChooser(self):
        self.toolboxGUIs.sort()
        self.toolboxchooser.update([tb.name() for tb in self.toolboxGUIs])
                
    #############################################

    def saveImage_gui(self, menuitem):
        filename = fileselector.fileSelector(
            title='Save Image', ident='save image')
        if filename:
            mode = fileselector.get_file_mode(filename, no_append=1)
            if mode:
                menuitem(filename=filename) # calls GhostGfxWindow.saveImage

    def saveContourmap_gui(self, menuitem):
        filename = fileselector.fileSelector(
            title='Save Contourmap', ident='save contourmap')
        if filename:
            mode = fileselector.get_file_mode(filename, no_append=1)
            if mode:
                menuitem(filename=filename)

    def backdate(self):
        # Backdate timestamps in a safe way.
        self.gfxlock.acquire()
        try:
            ghostgfxwindow.GhostGfxWindow.backdate(self)
        finally:
            self.gfxlock.release()
        
    def draw(self, *args):
        if BUFFERING:
            subthread.execute(self.draw_thread)
        else:
            mainthread.runBlock(self.draw_thread)
                
    def draw_thread(self):
        self.gfxlock.acquire()
        try:
            self.display.draw(self, self.device)
            # Zoom to full size if size hasn't yet been set.
            if self.realized and not self.zoomed:
                # Don't set self.zoomed here.  The zoom might not work.
                self.zoomFillWindow()
            # Update the contourmap info, now that the appropriate layer
            # has completed its draw.
            mainthread.runBlock(self.show_contourmap_info)
        finally:
            self.gfxlock.release()

            
    def toggleAntialias(self, menuitem, antialias):
        ghostgfxwindow.GhostGfxWindow.toggleAntialias(self, menuitem, antialias)
        self.newCanvas()
        self.draw()

    def zoomFactor(self):
        return self.settings.zoomfactor

    def zoomIn(self, *args):
        if self.oofcanvas and not self.oofcanvas.is_empty():
            mainthread.runBlock(self.oofcanvas.zoom,
                                (self.settings.zoomfactor,))
            self.zoomed = 1
    def zoomOut(self, *args):
        if self.oofcanvas and not self.oofcanvas.is_empty():
            mainthread.runBlock(self.oofcanvas.zoom,
                                (1./self.settings.zoomfactor,))
            self.zoomed = 1
    def zoomFillWindow(self, *args):
        mainthread.runBlock(self.zoomFillWindow_thread)
    def zoomFillWindow_thread(self):
        if self.oofcanvas and not self.oofcanvas.is_empty():
            bbox = self.oofcanvas.get_bounds()
##             width, height = self.oofcanvas.visible_size()
            width = self.oofcanvas.get_width_in_pixels()-1
            height = self.oofcanvas.get_height_in_pixels()-1
            if bbox.width() > 0 and bbox.height() > 0:
                xf = width/bbox.width()
                yf = height/bbox.height()
                ppu = min(xf, yf)       # pixels per unit
                self.oofcanvas.set_scrollregion(bbox, ppu)
                self.zoomed = 1

    def set_zoom(self, scrollreg, ppu):
        mainthread.runBlock(self.set_zoom_thread, (scrollreg, ppu))
    def set_zoom_thread(self, scrollreg, ppu):
        self.oofcanvas.set_scrollregion(scrollreg)
        self.oofcanvas.set_pixels_per_unit(ppu)
        self.zoomed = 1

    # GUI callback for new contourmap aspect ratio.
    # GUI callbacks are required because, when the settings change,
    # you have to redraw the window.
    def aspectRatio(self, menuitem, ratio):
        self.settings.aspectratio = ratio
        mainthread.run(self.show_contourmap_info)

    # Overridden menu callbacks for these settings.
    def contourmapMarkSize(self, menuitem, size):
        self.settings.contourmap_markersize = size
        v = self.contourmapdata.mark_value
        self.contourmapdata.mark_value = None
        mainthread.run(self.show_contourmap_ticks, (v,))

    def contourmapMarkColor(self, menuitem, color):
        self.settings.contourmap_markercolor = color
        v = self.contourmapdata.mark_value
        self.contourmapdata.mark_value = None
        mainthread.run(self.show_contourmap_ticks, (v,))

    def toggleLongLayerNames(self, menuitem, longlayernames):
        self.settings.longlayernames = longlayernames
        for layer in self.display:
            mainthread.runBlock(self.layerwidgets[layer].updateText)
        
    # Sets background color for both canvases.
    def bgColor(self, menuitem, color):          # OOFMenu callback
        ghostgfxwindow.GhostGfxWindow.bgColor(self, menuitem, color)
        mainthread.runBlock(self.oofcanvas.set_bgColor, (color,))
        mainthread.runBlock(self.contourmapdata.canvas.set_bgColor, (color,))

    def marginCB(self, menuitem, fraction):
        self.settings.margin = fraction
        mainthread.runBlock(self.oofcanvas.set_margin, (fraction,))
        
    ################################

    def makeLayerWidget(self, layer):
        if self.settings.listall or layer.listed:
            return layerwidget.ListedLayerWidget(self, layer)
        return layerwidget.UnListedLayerWidget(self, layer)

    def newLayer_gui(self, menuitem):
        layereditor.menu.LayerSet.New(window=self.name)

    def editLayer_gui(self, menuitem):
        layereditor.menu.LayerSet.Edit(
            window=self.name, layer_number=self.layerID(self.selectedLayer))

    def newLayer(self, displaylayer):
        mainthread.runBlock(self.newLayer_thread, (displaylayer,))
    def newLayer_thread(self, displaylayer):
        if len(self.display) > 0:
            self.layerwidgets[self.display[-1]].enableRaiseButton()
        lwidget = self.makeLayerWidget(displaylayer)
        lwidget.disableRaiseButton()
        self.layerList.prepend_items([lwidget.gtk])

        
    def deleteLayer_gui(self, menuitem):
        if self.selectedLayer is None:
            top().message_out('No layer is selected!\n')
        else:
            self.menu.Layer.Delete(
                n=self.layerID(self.selectedLayer))

    # This is an override of the command-line menu callback in
    # ghostgfxwindow.  It exists in addition to the GUI callback
    # below, which is also required because it operates on the
    # current layer.
    def hideLayer(self, menuitem, n):   # OOFMenu callback
        ghostgfxwindow.GhostGfxWindow.hideLayer(self, menuitem, n)
        mainthread.runBlock(self.hideLayer_thread, (menuitem, n))
    def hideLayer_thread(self, menuitem, n):
        self.display.hide_layer(self.device, n) # hide layer in canvas
        self.layerwidgets[self.display[n]].setShowButton(0) # set widget's btn
        # Update the contourmap.
        # self.contourmap_newlayers()
        self.show_contourmap_info()

        
    def hideLayer_gui(self, menuitem):  # OOFMenu GUI callback.
        if self.selectedLayer is None:
            reporter.report('No layer is selected!')
        else:
            self.menu.Layer.Hide(n=self.layerID(self.selectedLayer))

        
    def showLayer(self, menuitem, n):   # OOFMenu callback.
        ghostgfxwindow.GhostGfxWindow.showLayer(self, menuitem, n)
        mainthread.runBlock(self.showLayer_thread, (menuitem, n))
    def showLayer_thread(self, menuitem, n):
        self.display.show_layer(self.device, n) # show layer in canvas
        self.layerwidgets[self.display[n]].setShowButton(1)
        # Update the contourmap.
        # self.contourmap_newlayers()
        self.show_contourmap_info()


    def showLayer_gui(self, menuitem):  # OOFMenu GUI callback
        if self.selectedLayer is None:\
            reporter.report('No layer is selected!')
        else:
            self.menu.Layer.Show(n=self.layerID(self.selectedLayer))
    def layerHiddenByGUI(self, layer):  # called by LayerWidget
        self.menu.Layer.Hide(n=self.layerID(layer))
    def layerShownByGUI(self, layer):  # called by LayerWidget
        self.menu.Layer.Show(n=self.layerID(layer))

    def selectLayer(self, n):
        if self.selectedLayer is not None:
            self.deselectLayer(self.selectedLayer)
        ghostgfxwindow.GhostGfxWindow.selectLayer(self, n)
        mainthread.runBlock(self.layerwidgets[self.selectedLayer].select)
    def deselectLayer(self, n):
        if self.selectedLayer is not None:
            mainthread.runBlock(self.layerwidgets[self.selectedLayer].deselect)
            ghostgfxwindow.GhostGfxWindow.deselectLayer(self, n)
    def deselectAll(self):
        if self.selectedLayer is not None:
            mainthread.runBlock(self.layerwidgets[self.selectedLayer].deselect)
            ghostgfxwindow.GhostGfxWindow.deselectAll(self)
        
    def layerSelectedByGUI(self, layer): # called by LayerWidget
        self.menu.Layer.Select(n=self.layerID(layer))
    def layerDeselectedByGUI(self, layer): # called by LayerWidget
        if self.gtk:                # can occur after window is closed
            self.menu.Layer.Deselect(n=self.layerID(layer))


    # At layer-removal time, it's necessary to explicitly redraw the
    # contourmap, because the main canvas layers have not been redrawn,
    # so the usual automatic redraw has not taken place.
    def removeLayer(self, layer):
        ghostgfxwindow.GhostGfxWindow.removeLayer(self, layer)
        self.show_contourmap_info()
    
    def raiseLayer_gui(self, menuitem):
        if self.selectedLayer is None:
            top().message_out('No layer is selected!\n')
        else:
            self.menu.Layer.Raise.One_Level(
                n=self.layerID(self.selectedLayer))
    def raiseToTop_gui(self, menuitem):
        if self.selectedLayer is None:
            top().message_out('No layer is selected!\n')
        else:
            self.menu.Layer.Raise.To_Top(
                n=self.layerID(self.selectedLayer))
    def lowerLayer_gui(self, menuitem):
        if self.selectedLayer is None:
            top().message_out('No layer is selected!\n')
        else:
            self.menu.Layer.Lower.One_Level(
                n=self.layerID(self.selectedLayer))
    def lowerToBottom_gui(self, menuitem):
        if self.selectedLayer is None:
            top().message_out('No layer is selected!\n')
        else:
            self.menu.Layer.Lower.To_Bottom(
                n=self.layerID(self.selectedLayer))
    def raiseLayerByGUI(self, layer):
        self.menu.Layer.Raise.One_Level(n=self.layerID(layer))
    def lowerLayerByGUI(self, layer):
        self.menu.Layer.Lower.One_Level(n=self.layerID(layer))
        
    ##############################################

    def cloneWindow(self, *args):
        return mainthread.runBlock(self.cloneWindow_thread, args)
    def cloneWindow_thread(self, *args):
        ## This doesn't work completely correctly as far as copying sizes goes.
        # Create the clone
        clone = ghostgfxwindow.GhostGfxWindow.cloneWindow(self, *args)
        # Tell the clone what its size should be.
        gtkutils.copyWidgetSizes(self.gtk, clone.gtk)
        if not self.oofcanvas.is_empty():
            sr = self.oofcanvas.get_scrollregion()
            ppu = self.oofcanvas.get_pixels_per_unit()
            clone.set_zoom(sr, ppu)
            clone.oofcanvas.hScrollTo(self.oofcanvas.hScrollPosition())
            clone.oofcanvas.vScrollTo(self.oofcanvas.vScrollPosition())
        return clone


    ##### Scrolling ###########

    def hCanvasScrollCB(self, event, value):
        # called by RulerCanvas when the mouse is released after scrolling
        self.menu.Settings.Scroll.Horizontal(position=value)

    def vCanvasScrollCB(self, event, value):
        self.menu.Settings.Scroll.Vertical(position=value)

    def hScrollCB(self, menuitem, position): # OOFMenu callback
        ghostgfxwindow.GhostGfxWindow.hScrollCB(self, menuitem, position)
        mainthread.runBlock(self.oofcanvas.hScrollTo, (position,))

    def vScrollCB(self, menuitem, position): # OOFMenu callback
        ghostgfxwindow.GhostGfxWindow.vScrollCB(self, menuitem, position)
        mainthread.runBlock(self.oofcanvas.vScrollTo, (position,))

    #### Mouse clicks ############

    def setMouseHandler(self, handler):
        self.mouseHandler = handler

    def removeMouseHandler(self):
        self.mouseHandler = mousehandler.nullHandler

    def mouseCB(self, eventtype, x, y, shift, ctrl):
        debug.mainthreadTest()
        global _during_callback
        _during_callback = 1
        if self.mouseHandler.acceptEvent(eventtype):
            if eventtype == 'up':
                self.mouseHandler.up(x,y, shift, ctrl)
            elif eventtype == 'down':
                self.mouseHandler.down(x,y, shift, ctrl)
            elif eventtype == 'move':
                self.mouseHandler.move(x,y, shift, ctrl)
        _during_callback = 0

    def setRubberband(self, rubberband):
        self.rubberband = rubberband
        self.oofcanvas.set_rubberband(rubberband)

##############################################

# This function redefines the one in GfxWindowManager when the GUI
# code is loaded.

from SWIG.common.IO import threadstate

def _newWindow(self, name, **kwargs):
    if top(): # if in GUI mode
        return GfxWindow(name, self, **kwargs)
    return ghostgfxwindow.GhostGfxWindow(name, self, **kwargs)

gfxmanager.GfxWindowManager._newWindow = _newWindow

