# -*- python -*-
# $RCSfile: matrixparamwidgets.py,v $
# $Revision: 1.13 $
# $Author: langer $
# $Date: 2004/07/30 19:14:20 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# This file contains the definitions of input widgets used for
# the input of matrix data.  These are usable as parameterwidgets.

import gtk
import math
import types

from common import debug

from common.IO.GUI import parameterwidgets
from common.IO.GUI import widgetscope
from common.IO import parameter

# String for setting the width of the Matrix input fields in
# the current font.

_field_size_string = "0.000000"

# An array of FloatInput things.  Has as values a dictionary
# indexed by tuples of integers.
class MatrixInput(parameterwidgets.ParameterWidget,
                  widgetscope.WidgetScope):
    def __init__(self,label,rows,cols,value=None,scope=None):
        debug.mainthreadTest()
        parameterwidgets.ParameterWidget.__init__(
            self, gtk.GtkTable(rows=rows+2,cols=cols+2), scope)
        widgetscope.WidgetScope.__init__(self, scope)
        self.rows = rows
        self.cols = cols
        self.floats = {}
        #
        # self.gtk = gtk.GtkTable(rows=self.rows+2,cols=self.cols+2)
        #
        # Labels.
        for r in range(self.rows):
            lbl = gtk.GtkLabel(label+'('+`r+1`+',x)')
            self.gtk.attach(lbl,0,1,r+1,r+2)
        for c in range(self.cols):
            lbl = gtk.GtkLabel(label+'(x,'+`c+1`+')')
            self.gtk.attach(lbl,c+1,c+2,0,1)
        #
        for r in range(self.rows):
            for c in range(self.cols):
                # Parameters are quite lightweight, no harm in providing
                # a dummy to the FloatWidget.
                newfloat = parameterwidgets.FloatWidget(
                    parameter.FloatParameter('dummy', 0.0), scope=self)
                # Also, set the size -- otherwise, the matrix
                # array gets to be too big.
                font = newfloat.gtk.get_style().font
                width = font.width(_field_size_string)
                newfloat.gtk.set_usize(width,-1)
                #
                self.floats[(r,c)] = newfloat
                self.gtk.attach(newfloat.gtk,c+1,c+2,r+1,r+2,xoptions=gtk.FILL)

        self.setValidity(1)             # always valid

    #
    # Provides partial output if you provide it with an index.
    # With no index, it returns the full dictionary of values,
    # with an index, it returns the get_value of the indexed entry.
    #
    # It is the responsibility of the derived class to provide
    # get-value and set-value functionality.  Default functionality
    # is not (yet) provided.


#
# This class does different things on init than its parent,
# but is otherwise similar.
class SymmetricMatrixInput(MatrixInput):
    def __init__(self,label,rows,cols,value=None,scope=None):
        debug.mainthreadTest()
        parameterwidgets.ParameterWidget.__init__(
            self, gtk.GtkTable(rows=rows+2,cols=cols+2), scope)
        widgetscope.WidgetScope.__init__(self, scope)
        #scope.addWidget(self)
        self.rows = rows
        self.cols = cols
        self.floats = {}
        #
        #self.gtk = gtk.GtkTable(rows=self.rows+2,cols=self.cols+2)
        #
        # Do labels first.
        for r in range(self.rows):
            lbl = gtk.GtkLabel(label+'('+`r+1`+',x)') 
            self.gtk.attach(lbl,0,1,r+1,r+2)
        for c in range(self.cols):
            lbl = gtk.GtkLabel(label+'(x,'+`c+1`+')')
            self.gtk.attach(lbl,c+1,c+2,0,1)

        # Now put the actual floats in .
        for r in range(self.rows):
            for c in range(r,self.cols):
                newfloat = parameterwidgets.FloatWidget(
                    parameter.FloatParameter('dummy',0.0), scope=self)
                # Also, set the size -- otherwise, the matrix
                # array gets to be too big.
                font = newfloat.gtk.get_style().font
                width = font.width(_field_size_string)
                newfloat.gtk.set_usize(width,-1)
                
                self.floats[(r,c)] = newfloat
                try:
                    self.floats[(r,c)].set_value(value[(r,c)])
                    self.floats[(r,c)].set_position(0)
                except:
                    pass
                self.gtk.attach(newfloat.gtk,c+1,c+2,r+1,r+2,xoptions=gtk.FILL)
        self.setValidity(1)
