# -*- python -*-
# $RCSfile: quit.py,v $
# $Revision: 1.23 $
# $Author: langer $
# $Date: 2004/08/06 18:52:26 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 


from common import subthread
from common import debug
from common import excepthook
from common import guitop
from common import utils
from common.IO import threadmanager
from common.IO import mainmenu
from common import parallel_enable
from common.IO import threadmanager
from SWIG.common import switchboard
from SWIG.common import lock
import os
import sys



def quit():                             # called in text mode only
    debug.mainthreadTest()
    if mainmenu.OOF.logChanged() and not quiet():
        answer = utils.OOFeval('raw_input("*Save log file? [Yn]: ")')
        if answer in ('','Y','y','yes','Yes', 'YES'):
            mainmenu.OOF.File.Save.Python_Log()
    cleanup()
    reallyquit()

def reallyquit():
    excepthook.restore_excepthook()
    sys.exit()
    

quitlock = lock.Lock()
## This lock is to prevent other actions that are being executed on a
## separate thread, such as saving microstructures, to be left
## incomplete due to the quitting sequence. If such menuitem is being
## executed while quitting, quit will wait on THIS lock until it
## finishes.

def cleanup():
    quitlock.acquire()
    if parallel_enable.query() == 1:
        try:
            from SWIG.common import mpitools
            if mpitools.rank() == 0:
                mainmenu.OOF.LoadData.IPC.Quit() ## tells back-end to start quit process
        except ImportError:
            pass
    # Turn off logging, so that window closing, etc. won't be logged.
    mainmenu.OOF.haltLog()
    subthread.miniThreadManager.quit()
    threadmanager.threadManager.quit()
    if guitop.top():
        guitop.top().stop()
        guitop.top().destroy()
    
    # Let other components that want to do something shut down
    # cleanly.  This signal should be requested with
    # requestCallbackInThread to guarantee that all callbacks are
    # called before the main thread shuts down.
    switchboard.notify('quitting time')

_quiet = 0

def quiet():
    return _quiet

def set_quiet():
    global _quiet
    _quiet = 1

def disallow():
    quitlock.acquire()

def allow():
    quitlock.release()

