// -*- C++ -*-
// $RCSfile: fieldindex.C,v $
// $Revision: 1.6 $
// $Author: langer $
// $Date: 2004/06/28 14:53:28 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include "fieldindex.h"
#include "outputval.h"
#include "common/tostring.h"

std::vector<int> *ScalarFieldIndex::components() const {
  return new std::vector<int>;	// empty vector
}

void VectorFieldIndex::set(const std::vector<int> *component) {
  index_ = (*component)[0];
}

std::vector<int> *VectorFieldIndex::components() const {
  std::vector<int> *c = new std::vector<int>(1);
  (*c)[0] = index_;
  return c;
}


static const int rowset[] = { 0, 1, 2, 1, 0, 0 };
static const int colset[] = { 0, 1, 2, 2, 2, 1 };

int SymTensorIndex::row() const {
  return rowset[v];
}

int SymTensorIndex::col() const {
  return colset[v];
}

void SymTensorIndex::set(const std::vector<int> *component) {
  v = ij2voigt((*component)[0], (*component)[1]);
}

std::vector<int> *SymTensorIndex::components() const {
  std::vector<int> *c = new std::vector<int>(2);
  (*c)[0] = rowset[v];
  (*c)[1] = colset[v];
  return c;
}

int SymTensorIndex::ij2voigt(int i, int j) { // static function
  if(i == j)
    return i;
  return 6 - i - j;
}

int SymTensorIndex::str2voigt(const std::string &str) {	// static function
  // str must be "pq" where p and q are in ('x', 'y', 'z')
  return ij2voigt(str[0]-'x', str[1]-'x');
}

std::ostream &operator<<(std::ostream &os, const FieldIndex &fi) {
  fi.print(os);
  return os;
}

void ScalarFieldIndex::print(std::ostream &os) const {
  os << "ScalarFieldIndex()";
}

void VectorFieldIndex::print(std::ostream &os) const {
  os << "VectorFieldIndex(" << index_ << ")";
}

void SymTensorIndex::print(std::ostream &os) const {
  os << "SymTensorIndex(" << row() << "," << col() << ")";
}

std::ostream &operator<<(std::ostream &os, const IndexP &ip) {
  const FieldIndex &fi(ip);
  os << "IndexP(" << fi << ")";
}

IteratorP *getSymTensorIterator(Planarity planarity) {
  if(planarity == IN_PLANE)
    return new IteratorP(new SymTensorInPlaneIterator());
  if(planarity == OUT_OF_PLANE)
    return new IteratorP(new SymTensorOutOfPlaneIterator());
  return new IteratorP(new SymTensorIterator());
}

