# -*- python -*-
# $RCSfile: wrappers.py,v $
# $Revision: 1.16 $
# $Author: ericma $
# $Date: 2004/09/16 21:38:37 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

from SWIG.image import oofimage

from common import debug
from image import imagemodifier
from common import registeredclass
from common.IO import parameter
from SWIG.image.GRAINBDY import imageops
from SWIG.common import intarray

class FullGaborModifier(imagemodifier.ImageModifier):
    def __init__(self, a, b, numAngles,Threshold,Line_color,d,n,B,trimYN,t):
        self.a = a
        self.b = b
        self.numAngles=numAngles
        self.Threshold=Threshold
        self.Line_color=Line_color
        self.d=d
        self.n=n
        self.B=B
        self.trimYN=trimYN
        self.t=t       
    def __call__(self, image):
        dbls=imageops.grayify(image)
        for i in range(self.numAngles):
            phi = i*180./self.numAngles
            current = imageops.realGabor(dbls,self.a,self.b,phi)
            if i!=0:
                old=imageops.findLargerVals(current,old)
            else:
                old=current
        scaled = imageops.scaleArray(old, 0.0, 1.0,self.Line_color)
        bools=imageops.connect(scaled,self.Threshold,self.t,self.d,self.n,
                               self.B,self.trimYN)
        imageops.setFromBool(image,bools)      
        
registeredclass.Registration(
    'FullEdgeDetection',
    imagemodifier.ImageModifier,
    FullGaborModifier,
    ordering=100,
    params=[
    parameter.IntParameter('a',  value=3, tip="a in pixels"),
    parameter.IntParameter('b',  value=3, tip="b in pixels"),
    parameter.IntParameter('numAngles',value=6,
              tip="number of angles to apply the filter"),
    parameter.FloatParameter('Threshold', value=.5,
              tip="Threshold value used when thresholding image"),
    parameter.IntParameter('Line_color', value=2,
              tip="Colors of the lines to detect. (2) Both white and dark lines. (1) Only white lines (0) Only black lines."),
    parameter.IntParameter('d',value=7,
              tip="Value used for morphological closing"),
    parameter.IntParameter('n',value=9,
              tip="Number of interations"),
    parameter.IntParameter('B',value=5,
              tip="Half the size of the box in pixels"),
    parameter.IntParameter('trimYN',value=0,
              tip="Do you want to trim the image? [1 or 0]"),
    parameter.FloatParameter('t',value=.5,
              tip="Percent to decrease threshold by")
    ],
    tip="Apply a full edge detection process"
)

class RealGaborModifier(imagemodifier.ImageModifier):
    def __init__(self, a, b, numAngles,Threshold):
        self.a = a
        self.b = b
        self.numAngles=numAngles
        self.Threshold=Threshold
    def __call__(self, image):
        dbls = imageops.grayify(image)
        for i in range(self.numAngles): 
            phi = i*180./self.numAngles
            current = imageops.realGabor(dbls,self.a,self.b,phi)
            if i!=0:
                old=imageops.findLargerVals(current,old)
            else:
                old=current

        scaled = imageops.scaleArray2(old, 0.0, 1.0)
        imageops.setFromArray(image,scaled)
        
registeredclass.Registration(
    'RealGabor',
    imagemodifier.ImageModifier,
    RealGaborModifier,
    ordering=100,
    params=[
    parameter.IntParameter('a',  value=3, tip="a in pixels"),
    parameter.IntParameter('b',  value=3, tip="b in pixels"),
    parameter.IntParameter('numAngles',value=6,
                       tip="number of angles to apply the filter"),
    parameter.FloatParameter('Threshold', value=.5,
                       tip="Threshold value used when thresholding image")
    ],
    tip="Apply a real Gabor filter to detect edges"
)

class ImagGaborModifier(imagemodifier.ImageModifier):
    def __init__(self, a, b, numAngles, Threshold):
        self.a = a
        self.b = b
        self.numAngles=numAngles
        self.Threshold=Threshold       
    def __call__(self, image):
        dbls = imageops.grayify(image)
        for i in range(self.numAngles):
            phi = i*180./self.numAngles
            current = imageops.imagGabor(dbls,self.a,self.b,phi)
            if i!=0:
                old=imageops.findLargerVals(current,old)
            else:
                old=current           
        scaled = imageops.scaleArray2(old, 0.0, 1.0)
        imageops.setFromArray(image,scaled)        
     
registeredclass.Registration(
    'ImaginaryGabor',
    imagemodifier.ImageModifier,
    ImagGaborModifier,
    ordering=100,
    params=[
    parameter.IntParameter('a',  value=3, tip="a in pixels"),
    parameter.IntParameter('b',  value=3, tip="b in pixels"),
    parameter.IntParameter('numAngles',value=6,
                       tip="number of angles to apply the filter"),
    parameter.FloatParameter('Threshold', value=.5,
                       tip="Threshold value used when thresholding image")    
    ],
    tip="Apply an imaginary Gabor filter to detect edges"
)

class ModGaborModifier(imagemodifier.ImageModifier):
    def __init__(self, a, b, numAngles, Threshold):
        self.a = a
        self.b = b
        self.numAngles=numAngles
        self.Threshold=Threshold       
    def __call__(self, image):
        dbls = imageops.grayify(image)
        for i in range(self.numAngles):
            phi = i*180./self.numAngles
            current = imageops.modGabor(dbls,self.a,self.b,phi)
            if i!=0:
                old=imageops.findLargerVals(current,old)
            else:
                old=current          
        scaled = imageops.scaleArray2(old, 0.0, 1.0)
        imageops.setFromArray(image,scaled)        
        
registeredclass.Registration(
    'ModifiedGabor',
    imagemodifier.ImageModifier,
    ModGaborModifier,
    ordering=100,
    params=[
    parameter.IntParameter('a',  value=3, tip="a in pixels"),
    parameter.IntParameter('b',  value=3, tip="b in pixels"),
    parameter.IntParameter('numAngles',value=6,
                       tip="number of angles to apply the filter"),
    parameter.FloatParameter('Threshold', value=.5,
                       tip="Threshold value used when thresholding image")    
    ],
    tip="Apply a modified Gabor filter to detect edges"
)

class GaborModifier(imagemodifier.ImageModifier):
    def __init__(self, a, b, numAngles, Threshold):
        self.a = a
        self.b = b
        self.numAngles=numAngles
        self.Threshold=Threshold       
    def __call__(self, image):
        dbls=imageops.grayify(image)
        
        for i in range(self.numAngles):
            phi = i*180./self.numAngles
            current = imageops.normGabor(dbls,self.a,self.b,phi)
            if i!=0:
                old=imageops.findLargerVals(current,old)
            else:
                old=current
    
        scaled=imageops.scaleArray2(old,0.0,1.0)
        imageops.setFromArray(image,scaled)            
        
registeredclass.Registration(
    'NormalGabor',
    imagemodifier.ImageModifier,
    GaborModifier,
    ordering=100,
    params=[
    parameter.IntParameter('a',  value=3, tip="a in pixels"),
    parameter.IntParameter('b',  value=3, tip="b in pixels"),
    parameter.IntParameter('numAngles',value=6,
                       tip="number of angles to apply the filter"),
    parameter.FloatParameter('Threshold', value=.5,
                       tip="Threshold value used when thresholding image")    
    ],
    tip="Apply the normal Gabor filter to detect edges"
)

class NewGaborModifier(imagemodifier.ImageModifier):
    def __init__(self, a, b, numAngles,Threshold,Line_color):
        self.a = a
        self.b = b
        self.numAngles=numAngles
        self.Threshold=Threshold
        self.Line_color=Line_color
    def __call__(self, image):
##        dbls=imageops.grayify(image)
##        final = imageops.newGabor(dbls,self.a,self.b,0)
##        gradDirs=intarray.makeIntArray(image.sizeInPixels(),0)
##        for i in range(self.numAngles-1):
##            phi = (i+1)*180./self.numAngles
##            temp = imageops.newGabor(dbls,self.a,self.b,phi)
##            final = imageops.findLargerVals2(final,temp,gradDirs,i+1)
##        scaled = imageops.scaleArray(final,0.0,1.0,self.Line_color)
##        scaled = imageops.nonmaxSuppress(scaled,gradDirs);
##        imageops.setFromArray(image,scaled)
        
        dbls = imageops.grayify(image)
        for i in range(self.numAngles): 
            phi = i*180./self.numAngles
            current = imageops.newGabor(dbls,self.a,self.b,phi)
            if i!=0:
                old=imageops.findLargerVals(current,old)
            else:
                old=current

        scaled = imageops.scaleArray(old, 0.0, 1.0,self.Line_color)

        imageops.setFromArray(image,scaled)
        
registeredclass.Registration(
    'NewGabor',
    imagemodifier.ImageModifier,
    NewGaborModifier,
    ordering=100,
    params=[
    parameter.IntParameter('a',  value=3, tip="a in pixels"),
    parameter.IntParameter('b',  value=3, tip="b in pixels"),
    parameter.IntParameter('numAngles',value=4,
                       tip="number of angles to apply the filter"),
    parameter.FloatParameter('Threshold', value=.5,
                       tip="Threshold value used when thresholding image"),
    parameter.IntParameter('Line_color', value=2,
                        tip="Colors of the lines to detect. (2) Both white and dark lines. (1) Only white lines (0) Only black lines.")
    ],
    tip="Apply a new Gabor filter to detect edges"
)

class SobelModifier(imagemodifier.ImageModifier):
    def __call__(self, image):
        dbls = imageops.grayify(image)
        temp1=imageops.sobel(dbls,0)
        temp2=imageops.sobel(dbls,2)
        combined=imageops.combineVals(temp1,temp2)  
           
        scaled = imageops.scaleArray2(combined, 0.0, 1.0)

        imageops.setFromArray(image,scaled)
        
registeredclass.Registration(
    'Sobel',
    imagemodifier.ImageModifier,
    SobelModifier,
    ordering=200,
    params=[    ],
    tip="Apply the Sobel mask to detect edges"
)

class CannyModifier(imagemodifier.ImageModifier):
    def __init__(self, stdDev):
        self.stdDev=stdDev
    def __call__(self, image):
        dbls = imageops.grayify(image)
        largest=imageops.canny(dbls,self.stdDev)
 ##       smoothed = imageops.gaussSmooth(dbls,self.stdDev)
        
##        for i in range(4):
##            temp=imageops.sobel(smoothed,i)
##            if i!=0:
##                largest=imageops.findLargerVals(largest,temp)
##            else:
##                largest=temp 
           
        scaled = imageops.scaleArray2(largest, 0.0, 1.0)

        imageops.setFromArray(image,scaled)
        
registeredclass.Registration(
    'Canny',
    imagemodifier.ImageModifier,
    CannyModifier,
    ordering=200,
    params=[ parameter.FloatParameter('stdDev', value=1,
                       tip="Standard deviation for the gaussian smoothing")    ],
    tip="Apply the Canny Edge Detector to detect edges"
)

class GaussSmoothModifier(imagemodifier.ImageModifier):
    def __init__(self, stdDev):
        self.stdDev=stdDev
    def __call__(self, image):
        dbls = imageops.grayify(image)
        temp = imageops.gaussSmooth(dbls,self.stdDev)  
           
        scaled = imageops.scaleArray2(temp, 0.0, 1.0)

        imageops.setFromArray(image,scaled)
        
registeredclass.Registration(
    'GaussianSmoothing',
    imagemodifier.ImageModifier,
    GaussSmoothModifier,
    ordering=200,
    params=[ parameter.FloatParameter('stdDev', value=1,
                       tip="Standard deviation for the gaussian smoothing")  ],
    tip="Apply Gaussian Smoothing to a noisy image"
)

class LaplacianFilterModifier(imagemodifier.ImageModifier):
    def __call__(self, image):
        dbls = imageops.grayify(image)
        temp = imageops.laplacian(dbls)  
        scaled = imageops.scaleArray2(temp, 0.0, 1.0)
        imageops.setFromArray(image,scaled)
        
registeredclass.Registration(
    'LaplacianFilter',
    imagemodifier.ImageModifier,
    LaplacianFilterModifier,
    ordering=200,
    params=[ ],
    tip="Apply the Laplacian Filter to detect edges"
)

class LaplacianGaussFilterModifier(imagemodifier.ImageModifier):
    def __init__(self, stdDev):
        self.stdDev=stdDev
    def __call__(self, image):
        dbls = imageops.grayify(image)
        temp = imageops.laplacGauss(dbls,self.stdDev)  
        scaled = imageops.scaleArray2(temp, 0.0, 1.0)
        imageops.setFromArray(image,scaled)
        
registeredclass.Registration(
    'LaplacianGaussFilter',
    imagemodifier.ImageModifier,
    LaplacianGaussFilterModifier,
    ordering=200,
    params=[ parameter.FloatParameter('stdDev', value=1,
                       tip="Standard deviation for the filter") ],
    tip="Apply the Laplacian Filter to detect edges"
)

#####################################################
#functions that apply the filters at a single angle.#
#####################################################

##class RealGaborAngleModifier(imagemodifier.ImageModifier):
##    def __init__(self, a, b, phi):
##        self.a = a
##        self.b = b
##        self.phi = phi
##    def __call__(self, image):
##        dbls = imageops.grayify(image)
##        gaborimg = imageops.realGabor(dbls,self.a,self.b,self.phi)
##        scaled = imageops.scaleArray2(gaborimg, 0.0, 1.0)
##        imageops.setFromArray(image,scaled)
        
##registeredclass.Registration(
##    'RealGabor_Single_Angle',
##    imagemodifier.ImageModifier,
##    RealGaborAngleModifier,
##    ordering=100,
##    params=[
##    parameter.IntParameter('a',  value=3, tip="a in pixels"),
##    parameter.IntParameter('b',  value=3, tip="b in pixels"),
##    parameter.FloatParameter('phi',  value=0.0,
##                        tip="angle in degrees")
##    ],
##    tip="Apply a real Gabor filter at a given angle"
##)

##class ImagGaborAngleModifier(imagemodifier.ImageModifier):
##    def __init__(self, a, b, phi):
##        self.a = a
##        self.b = b    
##        self.phi = phi
##    def __call__(self, image):
##        dbls = imageops.grayify(image)
##        gaborimg = imageops.imagGabor(dbls,self.a,self.b,self.phi)
##        scaled = imageops.scaleArray2(gaborimg, 0.0, 1.0)
##        imageops.setFromArray(image,scaled)
    
##registeredclass.Registration(
##    'ImaginaryGabor_Single_Angle',
##    imagemodifier.ImageModifier,
##    ImagGaborAngleModifier,
##    ordering=100,
##    params=[
##    parameter.IntParameter('a',  value=3, tip="a in pixels"),
##    parameter.IntParameter('b',  value=3, tip="b in pixels"),   
##    parameter.FloatParameter('phi',  value=0.0,
##                        tip="angle in degrees")
##    ],
##    tip="Apply an imaginary Gabor filter at a given angle"
##)

##class ModGaborAngleModifier(imagemodifier.ImageModifier):
##    def __init__(self, a, b, phi):
##        self.a = a
##        self.b = b    
##        self.phi = phi
##    def __call__(self, image):
##        dbls = imageops.grayify(image)
##        gaborimg = imageops.modGabor(dbls,self.a,self.b,self.phi)
##        scaled = imageops.scaleArray2(gaborimg, 0.0, 1.0)
##        imageops.setFromArray(image,scaled)
        
##registeredclass.Registration(
##    'ModifiedGabor_Single_Angle',
##    imagemodifier.ImageModifier,
##    ModGaborAngleModifier,
##    ordering=100,
##    params=[
##    parameter.IntParameter('a',  value=3, tip="a in pixels"),
##    parameter.IntParameter('b',  value=3, tip="b in pixels"),   
##    parameter.FloatParameter('phi',  value=0.0,
##                        tip="angle in degrees")
##    ],
##    tip="Apply a modified Gabor filter to detect edges"
##)


##class GaborAngleModifier(imagemodifier.ImageModifier):
##    def __init__(self, a, b, phi):
##        self.a = a
##        self.b = b
##        self.phi
##    def __call__(self, image):
##        dbls = imageops.grayify(image)
##        gaborimg = imageops.modGabor(dbls,self.a,self.b,self.phi)
##        scaled = imageops.scaleArray(gaborimg, 0.0, 1.0)
##        imageops.setFromArray(image,scaled)
        
##registeredclass.Registration(
##    'NormalGabor',
##    imagemodifier.ImageModifier,
##    GaborAngleModifier,
##    ordering=100,
##    params=[
##    parameter.IntParameter('a',  value=3, tip="a in pixels"),
##    parameter.IntParameter('b',  value=3, tip="b in pixels"), 
##    parameter.FloatParameter('phi',  value=0.0,
##                        tip="angle in degrees")
##    ],
##    tip="Apply the normal Gabor filter to detect edges"
##)


class CloseImage(imagemodifier.ImageModifier):
    def __init__(self,n):
        self.n=n
    def __call__(self,image):
        dbls=imageops.grayify(image)
        bools=oofimage.threshold(dbls,.5)
        bools=imageops.close(bools,self.n)
        imageops.setFromBool(image,bools)

registeredclass.Registration(
    'CloseImage',
    imagemodifier.ImageModifier,
    CloseImage,
    ordering=10,
    params=[parameter.IntParameter('n',value=7,
                                tip="Diameter for closing. Must be odd")],
    tip="Close the gaps in a binary image"
)

class SkeletonizeImage(imagemodifier.ImageModifier):
    def __call__(self,image):
        dbls=imageops.grayify(image)
        bools=oofimage.threshold(dbls,.5)
        bools=imageops.skeletonize(bools)
        imageops.setFromBool(image,bools)

registeredclass.Registration(
    'SkeletonizeImage',
    imagemodifier.ImageModifier,
    SkeletonizeImage,
    ordering=11,
    tip="Skeletonize the image"
)

class ConnectEdges(imagemodifier.ImageModifier):
    def __init__(self,Threshold,d,n,B,trimYN,t):
        self.d=d;
        self.Threshold=Threshold
        self.n=n
        self.B=B
        self.trimYN=trimYN
        self.t=t
    def __call__(self,image):
        dbls=imageops.grayify(image)
 #     bools=oofimage.threshold(dbls,.5)
        bools=imageops.connect(dbls,self.Threshold,self.t,self.d,self.n,
                               self.B,self.trimYN)
        imageops.setFromBool(image,bools)

registeredclass.Registration(
    'Connect_Edges',
    imagemodifier.ImageModifier,
    ConnectEdges,
    ordering=11,
    params=[parameter.FloatParameter('Threshold',value=.5,
                                tip="Initial threshold value"),
            parameter.IntParameter('d',value=7,
                                tip="Value used for morphological closing"),
            parameter.IntParameter('n',value=9,
                                tip="Number of interations"),
            parameter.IntParameter('B',value=5,
                                tip="Half the size of the box in pixels"),
            parameter.IntParameter('trimYN',value=0,
                                tip="Do you want to trim the image? [1 or 0]"),
            parameter.FloatParameter('t',value=.5,
                                tip="Percent to decrease threshold by")],  
    tip="Connect the edges in the image"
)

class AddGausNoise(imagemodifier.ImageModifier):
    def __init__(self,Standard_deviation):
        self.Standard_deviation=Standard_deviation
    def __call__(self,image):
        dbls=imageops.grayify(image)
        dbls=imageops.addNoise(dbls,self.Standard_deviation)
        imageops.setFromArray(image,dbls)

registeredclass.Registration(
    "Add_Gaussian_Noise",
    imagemodifier.ImageModifier,
    AddGausNoise,
    ordering=200,
    params=[parameter.FloatParameter('Standard_deviation', value=.5)],
    tip="Adds Gausseian noise to an image"
)

class SpreadValues(imagemodifier.ImageModifier):
    def __init__(self,T):
        self.T=T
    def __call__(self,image):
        dbls=imageops.grayify(image)
        dbls=imageops.spread(dbls,self.T)
        imageops.setFromArray(image,dbls)

registeredclass.Registration(
    "SpreadDataValues",
    imagemodifier.ImageModifier,
    SpreadValues,
    ordering=200,
    params=[parameter.FloatRangeParameter('T', (0,1,.01),value=.5)],
    tip="Scale values over given threshold for larger spread."
)


class SpreadValues2(imagemodifier.ImageModifier):
    def __init__(self,T):
        self.T=T
    def __call__(self,image):
        dbls=imageops.grayify(image)
        dbls=imageops.spread2(dbls,self.T)
        imageops.setFromArray(image,dbls)

registeredclass.Registration(
    "SpreadDataValues2",
    imagemodifier.ImageModifier,
    SpreadValues2,
    ordering=200,
    params=[parameter.FloatRangeParameter('T', (0,1,.01),value=.5)],
    tip="Scale values under given threshold for larger spread."
)

class MakeHistogram(imagemodifier.ImageModifier):
    def __init__(self, val):
        self.val=val
    def __call__(self,image):
        dbls=imageops.grayify(image)
        imageops.printHistogram(dbls,self.val)

registeredclass.Registration(
    "MakeHistogram",
    imagemodifier.ImageModifier,
    MakeHistogram,
    ordering=300,
    params=[parameter.IntParameter('val',value=20)],
    tip="Creates a histogram"
)

class HysteresisThreshold(imagemodifier.ImageModifier):
    def __init__(self,T1,T2):
        self.T1=T1
        self.T2=T2
    def __call__(self, image):
        dbls = imageops.grayify(image)
        bools = imageops.hysteresisThresh(dbls,self.T1,self.T2)
        imageops.setFromBool(image,bools)
        
registeredclass.Registration(
    'HysteresisThreshold',
    imagemodifier.ImageModifier,
    HysteresisThreshold,
    ordering=100,
    params=[
    parameter.FloatRangeParameter('T1', (0,1,.01),value=.5,
                                  tip="Lower threshold value between 0 and 1"),
    parameter.FloatRangeParameter('T2', (0,1,.01),value=.5,
                                  tip="Higher threshold value between 0 and 1")
    ],
    tip="Threshold an image given a threshold value."
)

class compareImages(imagemodifier.ImageModifier):
    def __init__(self, val):
        self.val=val
    def __call__(self,image):
        dbls=imageops.grayify(image)
        bool1=oofimage.threshold(dbls,.5)
        imageops.compare(bool1,self.val)

registeredclass.Registration(
    "compareImages",
    imagemodifier.ImageModifier,
    compareImages,
    ordering=400,
    params=[parameter.IntParameter('val',value=3)],
    tip="Compare 2 images"
)
